#if !defined __SERIALMODEM_H

#define __SERIALMODEM_H

#include "SDL_net.h"
#include "serialport.h"
#include <assert.h>
#define MODEMSPD 57600
#define SREGS 100

//If it's too high you overflow terminal clients buffers i think
#define MODEM_BUFFER_QUEUE_SIZE 1024


struct ModemHd {
	char cmdbuf[MODEM_BUFFER_QUEUE_SIZE];
	bool commandmode;
	bool answermode;
	bool echo,response,numericresponse;
	bool telnetmode;
	Bitu cmdpause;
	Bits ringtimer;
	Bits ringcount;
	Bitu plusinc;
	Bitu cmdpos;

	Bit8u reg[SREGS];
	TCPsocket incomingsocket;
	TCPsocket socket; 
	TCPsocket listensocket;
	SDLNet_SocketSet socketset;

	IPaddress openip;

	Bitu comport;
	Bitu listenport;

	char remotestr[4096];

	bool dialing;
	double f1, f2;
	Bitu diallen;
	Bitu dialpos;
	char dialstr[256];
	// TODO: Re-enable dialtons
	//MIXER_Channel * chan;
};

enum ResTypes {
	ResNONE,
	ResOK,ResERROR,
	ResCONNECT,ResRING,
	ResBUSY,ResNODIALTONE,ResNOCARRIER,
};

#define TEL_CLIENT 0
#define TEL_SERVER 1

struct telnetClient {
	bool binary[2];
	bool echo[2];
	bool supressGA[2];
	bool timingMark[2];

	bool inIAC;
	bool recCommand;
	Bit8u command;
};

class CFifo {
public:
	CFifo(Bitu _size) {
		size=_size;
		pos=used=0;
		data=new Bit8u[size];
	}
	~CFifo() {
		delete[] data;
	}
	INLINE Bitu left(void) {
		return size-used;
	}
	INLINE Bitu inuse(void) {
		return used;
	}
	void clear(void) {
		used=pos=0;
	}
	bool isFull() {
		return (used >= size);
	}
	void addb(Bit8u _val) {
		assert(used<size);
		Bitu where=pos+used;
		if (where>=size) where-=size;
		data[where]=_val;
		used++;
	}
	void adds(Bit8u * _str,Bitu _len) {
		assert((used+_len)<=size);
		Bitu where=pos+used;
		used+=_len;
		while (_len--) {
			if (where>=size) where-=size;
			data[where++]=*_str++;
		}
	}
	Bit8u getb(void) {
		if (!used) return data[pos];
		Bitu where=pos;
		if (++pos>=size) pos-=size;
		used--;
		return data[where];
	}
	void gets(Bit8u * _str,Bitu _len) {
		assert(used>=_len);
		used-=_len;
		while (_len--) {
			*_str++=data[pos];
			if (++pos>=size) pos-=size;
		}
	}
private:
	Bit8u * data;
	Bitu size,pos,used;
	Bit8u tmpbuf[MODEM_BUFFER_QUEUE_SIZE];
	
	
};

class CSerialModem : public CSerial {
public:

	CFifo *rqueue;
	CFifo *tqueue;
	ModemHd mhd;

	//bool rxempty;
	//Bit8u mctrl;

	CSerialModem(Bit16u baseAddr,
		Bit8u initIrq,
		Bit32u initBps,
		Bit8u bytesize,
		const char* parity,
		Bit8u stopbits,

		const char *remotestr = NULL,
		Bit16u lport = 23);

	~CSerialModem();

	void SendLine(const char *line);

	void SendRes(ResTypes response);

	

	//void Send(Bit8u val) {
	//	tqueue->addb(val);
	//}

	//Bit8u Recv(Bit8u val) {
	//	return rqueue->getb();
//
	//}

	void openConnection(void);
	void updatemstatus(void);
	bool Dial(char * host);
	void AcceptIncomingCall(void);
	Bitu ScanNumber(char * & scan);
	void HangUp(void);

	void DoCommand();
	
	void MC_Changed(Bitu new_mc);

	void TelnetEmulation(Bit8u * data, Bitu size);

	void Timer2(void);

void RXBufferEmpty();

void transmitByte(Bit8u val);
void updatePortConfig(Bit8u dll, Bit8u dlm, Bit8u lcr);
void updateMSR();

void setBreak(bool);

void updateModemControlLines(Bit8u mcr);
bool CanSend(void);
bool CanRecv(void);


protected:
	char cmdbuf[MODEM_BUFFER_QUEUE_SIZE];
	bool commandmode;
	bool answermode;
	bool echo;
	bool telnetmode;
	Bitu cmdpause;
	Bits ringtimer;
	Bits ringcount;
	Bitu plusinc;
	Bitu cmdpos;
		
	Bit8u reg[SREGS];
	IPaddress openip;
	TCPsocket incomingsocket;
	TCPsocket socket;
	TCPsocket listensocket;
	SDLNet_SocketSet socketset;

	struct {
		bool binary[2];
		bool echo[2];
		bool supressGA[2];
		bool timingMark[2];
					
		bool inIAC;
		bool recCommand;
		Bit8u command;
	} telClient;
	struct {
		bool active;
		double f1, f2;
		Bitu len,pos;
		char str[256];
	} dial;
};

#endif 
