/*
 *  Copyright (C) 2002-2005  The DOSBox Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <assert.h>
#include <list>

#include "dosbox.h"
#include "serialport.h"
#include "ds.h"
#include "serialdummy.h"
#include "modem.h"


#include "inout.h"
//#include "mixer.h"
#include "pic.h"
#include "setup.h"
#include "timer.h"
//#include "math.h"
//#include "regs.h"




//#define SERIALBASERATE 115200
#define LOG_UART LOG_MSG

CSerialList seriallist;

static bool installTimerHandler = true;


void CSerial::Timer(void)
{
	////LOG_UART("Serial port at %x: Timer", base);
	if(loopback_pending)
	{
		//LOG_UART("Serial port at %x: Loopback sent back", base);
		
		loopback_pending = false;
		receiveByte(loopback_data);
		ByteTransmitted();
	}
}

void SERIAL_Update(void)
{
	for (CSerial_it it=seriallist.begin();it!=seriallist.end();it++)
	{
		CSerial * serial=(*it);
		serial->Timer();
		serial->Timer2();
	}
}

static void WriteSerial(Bitu port,Bitu val,Bitu iolen) {
	Bitu check=port&~0x7;
	for (CSerial_it it=seriallist.begin();it!=seriallist.end();it++){
		CSerial * serial=(*it);
		if (check==serial->base)
		{
			switch(port&0x7)
			{
			case THR_OFFSET:
				//LOG_UART("%x: write THR, value 0x%x", port, val);
				serial->Write_THR(val);
				return;
			
			case IER_OFFSET:
				//LOG_UART("%x: write IER, value 0x%x", port, val);
				serial->Write_IER(val);
				return;
			
				
			case LCR_OFFSET:
				//LOG_UART("%x: write LCR, value 0x%x", port, val);
				serial->Write_LCR(val);
				
				return;
			case MCR_OFFSET:

				//LOG_UART("%x: write MCR, value 0x%x", port, val);
				serial->Write_MCR(val);
				return;

			case MSR_OFFSET:
				//LOG_UART("%x: write MSR, value 0x%x", port, val);
				serial->Write_MSR(val);
				return;
				
			case SPR_OFFSET:
				//LOG_UART("%x: write SPR, value 0x%x", port, val);
				serial->Write_SPR(val);
				return;
			default:
					serial->Write_reserved(val,port&0x7);
			}
		}
	}
}

static Bitu ReadSerial(Bitu port,Bitu iolen) {
	Bitu check=port&~0x7;
	
	for (CSerial_it it=seriallist.begin();it!=seriallist.end();it++){
		CSerial * serial=(*it);
		if (check==serial->base)
		{
				switch(port&0x7)
				{
				case RHR_OFFSET: 
					//LOG_UART("%x: read RHR",port);
					return serial->Read_RHR();
				case IER_OFFSET: 
					//LOG_UART("%x: read IER",port);
					return serial->Read_IER();
				case ISR_OFFSET: 
					//LOG_UART("%x: read ISR",port);
					return serial->Read_ISR();
				case LCR_OFFSET: 
					//LOG_UART("%x: read LCR",port);
					return serial->Read_LCR();
				case MCR_OFFSET: 
					//LOG_UART("%x: read MCR",port);
					return serial->Read_MCR();
				case LSR_OFFSET: 
					//LOG_UART("%x: read LSR",port);
					return serial->Read_LSR();
				case MSR_OFFSET: 
					//LOG_UART("%x: read MSR",port);
					return serial->Read_MSR();
				case SPR_OFFSET: 
					//LOG_UART("%x: read SPR",port);
					return serial->Read_SPR();
				}
		}
	}
	return 0;
}



/*****************************************************************************/
/* Interrupt control routines												**/
/*****************************************************************************/
_inline void CSerial::rise(Bit8u priority)
{
	//LOG_UART("Serial port at %x: Rise priority 0x%x", base, priority);
	waiting_interrupts|= priority;
	WriteRealIER(IER);
}

// clears the pending interrupt, triggers other waiting interrupt
_inline void CSerial::clear(Bit8u priority)
{
	//LOG_UART("Serial port at %x: cleared priority 0x%x", base, priority);
	waiting_interrupts&= (~priority);
	WriteRealIER(IER);
}

void CSerial::WriteRealIER(Bit8u data)
{
	//LOG_UART("Serial port at %d: RealIER, value 0x%x", base, data);
	data=data&0xF;	// THE UPPER ONES ALWAYS READ 0! NOTHIN' ELSE!
	Bit8u old_pending_interrupts = pending_interrupts;
	Bit8u difference_pending_interrupts;
	
	// rise TX AGAIN when present and is being enabled
	if((data&TX_PRIORITY)&&(!(IER&TX_PRIORITY)))
		if(LSR&LSR_TX_HOLDING_EMPTY_MASK)
			waiting_interrupts|=TX_PRIORITY;

	pending_interrupts = waiting_interrupts&data;
	if((difference_pending_interrupts = (pending_interrupts ^ old_pending_interrupts)))
	{	// something in pending interrupts has changed
		if(difference_pending_interrupts & pending_interrupts)
		{	// some new bits were set
			if(!current_priority)
			{	// activate interrupt..
				if(pending_interrupts & ERROR_PRIORITY)
				{
					current_priority = ERROR_PRIORITY;
					ISR= ISR_ERROR_VAL;
				}
				else if(pending_interrupts & RX_PRIORITY)
				{
					current_priority = RX_PRIORITY;
					ISR= ISR_RX_VAL;
				}
				else if(pending_interrupts & TX_PRIORITY)
				{
					current_priority = TX_PRIORITY;
					ISR= ISR_TX_VAL;
				}
				else if(pending_interrupts & MSR_PRIORITY)
				{
					current_priority = MSR_PRIORITY;
					ISR= ISR_MSR_VAL;
				}
				//LOG_UART("Serial port at %x: Interrupt activated: priority %d", base,current_priority);
				PIC_ActivateIRQ(irq);

			} // else the new interrupts were already
			// written into pending_interrupts
		} // else no new bits were set

		if(difference_pending_interrupts&(~pending_interrupts)) // some bits were reset
		{
			if(pending_interrupts)
			{	// some more are waiting
				if(!(current_priority&pending_interrupts))
				{	// the current interrupt has been cleared
					// choose the next one
					if(pending_interrupts & ERROR_PRIORITY)
					{
						current_priority = ERROR_PRIORITY;
						ISR= ISR_ERROR_VAL;
					}
					else if(pending_interrupts & RX_PRIORITY)
					{
						current_priority = RX_PRIORITY;
						ISR= ISR_RX_VAL;
					}
					else if(pending_interrupts & TX_PRIORITY)
					{
						current_priority = TX_PRIORITY;
						ISR= ISR_TX_VAL;
					}
					else if(pending_interrupts & MSR_PRIORITY)
					{
						current_priority = MSR_PRIORITY;
						ISR= ISR_MSR_VAL;
					}
				}
			}
			else
			{	
				current_priority = NONE_PRIORITY;
				ISR= ISR_CLEAR_VAL;
				PIC_DeActivateIRQ(irq);
				//LOG_UART("Serial port at %x: Interrupt deactivated.", base);
			}
		}
	}
	IER = data;
}



/*****************************************************************************/
/* Internal register modification											**/
/*****************************************************************************/
void CSerial::changeMSR(Bit8u data)
{
	if(!(MCR&MCR_LOOPBACK_Enable_MASK))
	{
		Bit8u temp=MSR;
		// look for signal changes
		if(temp|=((data&MSR_LINE_MASK)^(MSR&MSR_LINE_MASK))>>4)
		{	// some signals changed
			temp&=MSR_delta_MASK;// clear line states
			temp|=(data&MSR_LINE_MASK);	// set new line states
			
			MSR=temp;
			rise(MSR_PRIORITY);
		}
		// else nothing happened

		
	}
}
void CSerial::changeMSR_Loopback(Bit8u data)
{
	Bit8u temp=MSR;
	// look for signal changes
	if(temp|=((data&MSR_LINE_MASK)^(MSR&MSR_LINE_MASK))>>4)
	{	// some signals changed
		temp&=MSR_delta_MASK;// clear line states
		temp|=(data&MSR_LINE_MASK);	// set new line states
		
		MSR=temp;
		rise(MSR_PRIORITY);
	}
	// else nothing happened
}

/*****************************************************************************/
/* Can a byte be received?													**/
/*****************************************************************************/
bool CSerial::CanReceiveByte()
{
	return !(LSR&LSR_RX_DATA_READY_MASK);

}




/*****************************************************************************/
/* A byte was received														**/
/*****************************************************************************/
	void CSerial::receiveByte(Bit8u data)
{
	//LOG_UART("Serial port at 0x%x: byte received: %d", base,data);

	if(LSR&LSR_RX_DATA_READY_MASK)	// Overrun error ;o
	{
		LOG_UART("Serial port at 0x%x: RX Overrun!", base, data);
		LSR |= LSR_OVERRUN_ERROR_MASK;
		rise(ERROR_PRIORITY);
	}
	else
	{
        RHR = data;
		LSR|= LSR_RX_DATA_READY_MASK;
		rise(RX_PRIORITY);
	}
}

/*****************************************************************************/
/* A line error was received												**/
/*****************************************************************************/
	void CSerial::receiveError(Bit8u errorword)
{
	// LOG_UART("Serial port at 0x%x: RX Overrun!", base, data);
	// LSR |= LSR_OVERRUN_ERROR_MASK;
	LSR |= errorword;

	rise(ERROR_PRIORITY);
}

/*****************************************************************************/
/* ByteTransmitted: When a byte was sent, notify here.								**/
/*****************************************************************************/
void CSerial::ByteTransmitted()
{
	if(LSR&LSR_TX_HOLDING_EMPTY_MASK)
		{	// one space was empty
			// now both are
			LSR|=LSR_TX_EMPTY_MASK;
		}
	else	// both were full, now 1 is empty
	{
		if(MCR&MCR_LOOPBACK_Enable_MASK)
		{	// loopback mode
			LSR&=LSR_TX_HOLDING_EMPTY_MASK;
			transmitLoopbackByte(THR);
			rise(TX_PRIORITY);
		}
		else
		{	// direct "real" mode
			transmitByte(THR);
			rise(TX_PRIORITY);
		}
		
	}

}

/*****************************************************************************/
/* Transmit Holding Register, also LSB of Divisor Latch (r/w)				**/
/*****************************************************************************/
void CSerial::Write_THR(Bit8u data)
{	
	
	if((LCR & LCR_DIVISOR_Enable_MASK))
	{	
		//LOG_UART("Serial port at %x: Write to DLL value 0x%x", base, data);
		// write to DLL
		DLL = data;
		updatePortConfig(DLL, DLM, LCR);
	}
	else
	{
		//LOG_UART("Serial port at %x: Write to THR value 0x%x", base, data);
		// write to THR
		clear(TX_PRIORITY);
		
		if(LSR&LSR_TX_HOLDING_EMPTY_MASK)
		{	// one is empty	
			if(LSR&LSR_TX_EMPTY_MASK)
			{	// both are empty
				
				if(MCR&MCR_LOOPBACK_Enable_MASK)
				{	// loopback mode
					LSR&=(~LSR_TX_EMPTY_MASK);
					transmitLoopbackByte(data);
				}
				else
				{	
					// direct "real" mode
					LSR&=(~LSR_TX_EMPTY_MASK);
					
					transmitByte(data);
				}
				rise(TX_PRIORITY);
			}
			else 
			{	// only THR is empty; add byte and clear holding bit
				LSR&=(~LSR_TX_HOLDING_EMPTY_MASK);//&0x7f;
				THR = data;
			}
		}
		else
		{
			//LOG_UART("Serial port at %x: TX Overflow!", base);
			// TX overflow; how does real hardware respond to that... I do nothing
		}
	}
}


/*****************************************************************************/
/* Receive Holding Register, also LSB of Divisor Latch (r/w)				**/
/*****************************************************************************/
Bitu CSerial::Read_RHR()// Receive Holding Register, also LSB of Divisor Latch (r/w)
{	
	if((LCR & LCR_DIVISOR_Enable_MASK))
	{
		////LOG_UART("Serial port at %d: Read from DLL value 0x%x", base, DLL);
		return DLL;
	}
	else
	{
		clear(RX_PRIORITY);
		LSR&=(~LSR_RX_DATA_READY_MASK);
		//LOG_UART("Serial port at %d: Read from RHR value 0x%x", base, RHR);
		RXBufferEmpty();
		return RHR;
	}
}

/*****************************************************************************/
/* Interrupt Enable Register, also MSB of Divisor Latch (r/w)				**/
/*****************************************************************************/
// Modified by:
// - writing to it.
Bitu CSerial::Read_IER()
{
	
	if((LCR & LCR_DIVISOR_Enable_MASK))// IER or MSB?
	{
		//LOG_UART("Serial port at %d: Read from DLM value 0x%x", base, DLM);
		return DLM;
	}
	else
	{
		//LOG_UART("Serial port at %d: Read from IER value 0x%x", base, IER);
		return IER;
	}
}

void CSerial::Write_IER(Bit8u data)
{	
	if((LCR & LCR_DIVISOR_Enable_MASK))
	{	// write to DLM
		//LOG_UART("Serial port at %d: Write to DLM value 0x%x", base, data);
		DLM = data;
		updatePortConfig(DLL, DLM, LCR);
	}
	else
	{
		//LOG_UART("Serial port at %d: Write to IER value 0x%x", base, data);
		WriteRealIER(data);
		
	}
}
/*****************************************************************************/
/* Interrupt Status Register (r)											**/
/*****************************************************************************/
// modified by:
// - incoming interrupts
// - loopback mode
Bitu CSerial::Read_ISR()// Interrupt Status Register
{	
	Bit8u retval = ISR;
	//LOG_UART("Serial port at %d: Read from ISR value 0x%x", base, retval);
	
	// clear changes ISR!! mean..
	clear(TX_PRIORITY);
	return retval;
}
/*****************************************************************************/
/* Line Control Register (r/w)												**/
/*****************************************************************************/
// signal decoder configuration:
// - parity, stopbits, word length
// - send break
// - switch between RHR/THR and baud rate registers
// Modified by:
// - writing to it.
Bitu CSerial::Read_LCR()
{	
	
	//LOG_UART("Serial port at %d: Read from LCR value 0x%x", base, LCR);
	return LCR;
}
void CSerial::Write_LCR(Bit8u data)
{	
	Bit8u lcr_old = LCR;
	LCR = data;
	
	//LOG_UART("Serial port at %x: Write to LCR value 0x%x", base, data);
	
	// for debug output
	if(!(lcr_old & LCR_DIVISOR_Enable_MASK))
	{
		//if(data & LCR_DIVISOR_Enable_MASK)
			//LOG_UART("Serial port at 0x%x: Divisor-mode entered", base);
	}
	else
	{
		//if(!(data & LCR_DIVISOR_Enable_MASK))
			//LOG_UART("Serial port at 0x%x: Divisor-mode exited", base);
	}
	if((data^lcr_old)&LCR_PORTCONFIG_MASK)
	{
		//LOG_UART("Serial port at 0x%x: comm parameters changed", base);
		updatePortConfig(DLL, DLM, LCR);
	}
	
	if((LCR&LCR_BREAK_MASK)^(lcr_old&LCR_BREAK_MASK))
	{
		if(LCR&LCR_BREAK_MASK)setBreak(true);
		else setBreak(false);
	}
}
/*****************************************************************************/
/* Modem Control Register (r/w)												**/
/*****************************************************************************/
// Set levels of RTS and DTR, as well as loopback-mode.
// Modified by: 
// - writing to it.
Bitu CSerial::Read_MCR()
{	
	
	//LOG_UART("Serial port at %d: Read from MCR value 0x%x", base, MCR);
	return MCR;
}
void CSerial::Write_MCR(Bit8u data)
{	
	data&=0x1F;	// UPPER BITS ALWAYS 0!!!
	
	//LOG_UART("Serial port at %d: Write to MCR value 0x%x", base, data);
	
	if(MCR&MCR_LOOPBACK_Enable_MASK)
		if(data&MCR_LOOPBACK_Enable_MASK)
		{	// was on, is now on
			Bit8u param = 0;
			
			// RTS->CD
			// DTR->RI
			// OP1->DSR
			// OP2->CTS

			if(data&MCR_RTS_MASK) param|=MSR_CD_MASK;
			if(data&MCR_DTR_MASK) param|=MSR_RI_MASK;
			if(data&MCR_OP1_MASK) param|=MSR_DSR_MASK;
			if(data&MCR_OP2_MASK) param|=MSR_CTS_MASK;

			changeMSR_Loopback(param);

		}
		else
		{
			updateModemControlLines(data);
			// is switched off now
		}
	else
	{
		if(data&MCR_LOOPBACK_Enable_MASK)
		{	// is switched on:
			Bit8u par = 0;
			if(data&MCR_RTS_MASK) par|=MSR_CD_MASK;
			if(data&MCR_DTR_MASK) par|=MSR_RI_MASK;
			if(data&MCR_OP1_MASK) par|=MSR_DSR_MASK;
			if(data&MCR_OP2_MASK) par|=MSR_CTS_MASK;

			changeMSR(par);
			
			// go back to empty state
			LSR&=(LSR_TX_EMPTY_MASK|LSR_TX_HOLDING_EMPTY_MASK);

		}
		else
		{
			updateModemControlLines(data);
			// loopback is off
		}
	}
	MCR = data;
}

/*****************************************************************************/
/* Line Status Register (r)													**/
/*****************************************************************************/
// errors, tx registers status, rx register status
// modified by:
// - event from real serial port
// - loopback
Bitu CSerial::Read_LSR()
{
	Bit8u x = 0;
	x++;
	//x=~LSR_ERROR_MASK;
	Bitu retval = LSR;
	LSR&=(~LSR_ERROR_MASK);	// clear error bits on read
	//LOG_UART("Serial port at %d: Read from LSR value 0x%x", base, retval);
	clear(ERROR_PRIORITY);
	return retval;
}

void CSerial::Write_MSR(Bit8u val)
{
	MSR &= MSR_LINE_MASK;
	MSR |= val&MSR_delta_MASK;
}

/*****************************************************************************/
/* Modem Status Register (r)												**/
/*****************************************************************************/
// Contains status of the control input lines (CD, RI, DSR, CTS) and
// their "deltas": if level changed since last read delta = 1.
// modified by:
// - real values
// - write operation to MCR in loopback mode
Bitu CSerial::Read_MSR()
{	
	Bit8u retval = 0;
	if(!(MCR&MCR_LOOPBACK_Enable_MASK))
	{
		updateMSR();
	}
	retval = MSR;
	//LOG_UART("Serial port at %d: Read from MSR value 0x%x", base, MSR);
	clear(MSR_PRIORITY);
	// clear deltas
	MSR&=MSR_LINE_MASK;
	return retval;
}

/*****************************************************************************/
/* Scratchpad Register (r/w)												**/
/*****************************************************************************/
// Just a memory register. Not much to do here.
Bitu CSerial::Read_SPR() 
{	
	
	//LOG_UART("Serial port at %d: Read from SPR value 0x%x", base, SPR);
	return SPR;
}
void CSerial::Write_SPR(Bit8u data)
{	
	//LOG_UART("Serial port at %d: Write to SPR value 0x%x", base, data);
	SPR = data;
}
/*****************************************************************************/
/* Write_reserved															**/
/*****************************************************************************/
void CSerial::Write_reserved(Bit8u data, Bit8u address)
{	
	LOG_UART("Serial port at 0x%x: Write to reserved register, value 0x%x, register %x", base, data,address);
	
}

/*****************************************************************************/
/* Loopback: add byte to loopback buffer; it is received next time Timer	**/
/* function is invoked; (time needed is not emulated correctly, but I don't	**/
/* think this is sooooo important....)										**/
/*****************************************************************************/
void CSerial::transmitLoopbackByte(Bit8u val)
{
	//LOG_MSG("Serial port at %x: Loopback requested", base);
	loopback_pending = true;
	loopback_data = val;

}

/*****************************************************************************/
/* MSR Access																**/
/*****************************************************************************/
bool CSerial::getRI()
{
	if((MSR&MSR_RI_MASK)==0) return false;
	else return true;
}
bool CSerial::getCD()
{
	if((MSR&MSR_CD_MASK)==0) return false;
	else return true;
}
bool CSerial::getDSR()
{
	if((MSR&MSR_DSR_MASK)==0) return false;
	else return true;
}
bool CSerial::getCTS()
{
	if((MSR&MSR_CTS_MASK)==0) return false;
	else return true;
}

void CSerial::setRI(bool value)
{
	bool compare;
	if(MSR&MSR_RI_MASK) compare = true;
	else compare = false;
	if(value!=compare)
	{
		if(value)MSR|=MSR_RI_MASK;
		else MSR&=(~MSR_RI_MASK);
		MSR|=MSR_dRI_MASK;
		rise(MSR_PRIORITY);
	}
	//else no change
}
void CSerial::setDSR(bool value)
{
	bool compare;
	if(MSR&MSR_DSR_MASK) compare = true;
	else compare = false;
	if(value!=compare)
	{
		if(value)MSR|=MSR_DSR_MASK;
		else MSR&=(~MSR_DSR_MASK);
		MSR|=MSR_dDSR_MASK;
		rise(MSR_PRIORITY);
	}
	//else no change
}
void CSerial::setCD(bool value)
{
	bool compare;
	if(MSR&MSR_CD_MASK) compare = true;
	else compare = false;
	if(value!=compare)
	{
		if(value)MSR|=MSR_CD_MASK;
		else MSR&=(~MSR_CD_MASK);
		MSR|=MSR_dCD_MASK;
		rise(MSR_PRIORITY);
	}
	//else no change
}
void CSerial::setCTS(bool value)
{
	bool compare;
	if(MSR&MSR_CTS_MASK) compare = true;
	else compare = false;
	if(value!=compare)
	{
		if(value)MSR|=MSR_CTS_MASK;
		else MSR&=(~MSR_CTS_MASK);
		MSR|=MSR_dCTS_MASK;
		rise(MSR_PRIORITY);
	}
	//else no change
}

/*****************************************************************************/
/* Initialisation															**/
/*****************************************************************************/
void CSerial::Init_Registers(Bit32u initbps,
				  Bit8u bytesize, const char* parity, Bit8u stopbits)
{
	Bit8u lcrresult=0;
	Bit16u baudresult=0;
	
	RHR = 0;	
	THR = 0;	
	IER = 0;
	ISR = 0x1;
	LCR = 0;
	MCR = 0;
	LSR = 0x60;
	MSR = 0;

	SPR = 0xFF;

	DLL = 0x0;
	DLM = 0x0;

	pending_interrupts = 0x0;	
	current_priority = 0x0;
	waiting_interrupts = 0x0;	
	loopback_pending = false;


	// make lcr: byte size, parity, stopbits, baudrate
	
	if(bytesize==5) lcrresult|=LCR_DATABITS_5;
	else if(bytesize==6) lcrresult|=LCR_DATABITS_6;
	else if(bytesize==7) lcrresult|=LCR_DATABITS_7;
	else lcrresult|=LCR_DATABITS_8;

	if (parity[0] == 'O'||parity[0] == 'o')	lcrresult|= LCR_PARITY_ODD;
	else if (parity[0] == 'E'||parity[0] == 'e')	lcrresult|= LCR_PARITY_EVEN;
	else if (parity[0] == 'M'||parity[0] == 'm')	lcrresult|= LCR_PARITY_MARK;
	else if (parity[0] == 'S'||parity[0] == 's')	lcrresult|= LCR_PARITY_SPACE;
	else lcrresult|= LCR_PARITY_NONE;
	
	if(stopbits==2) lcrresult |= LCR_STOPBITS_MORE_THAN_1;	
	else lcrresult |= LCR_STOPBITS_1;

	// baudrate
	
	if(initbps>0) baudresult = (Bit16u)(115200 / initbps);
	else baudresult=12; // = 9600 baud

	Write_LCR(LCR_DIVISOR_Enable_MASK);
	Write_THR((Bit8u)baudresult&0xff);
	Write_IER((Bit8u)(baudresult>>8));
	Write_LCR(lcrresult);

}

CSerial::CSerial (Bit16u initbase, Bit8u initirq, Bit32u initbps,
				  Bit8u bytesize, const char* parity, Bit8u stopbits)
{

	Bitu i;
	//Bit16u initdiv;
	
	//Init_Registers();
	if(installTimerHandler)	// Gee, I wish FastDelegates were used..
	{
		TIMER_AddTickHandler(&SERIAL_Update);
		installTimerHandler =false;
	}
	base=initbase;
	irq=initirq;

	// Register I/O-Ports

	for (i=base;i<=(base+8);i++)
	{
		IO_RegisterWriteHandler(i,WriteSerial,IO_MB);
		IO_RegisterReadHandler(i,ReadSerial,IO_MB);
	}

};

CSerial::~CSerial(void)
{

};




void initcomport(Section* sec, int comnumber)
{
	const char* typestring;
	Bit8u irq;
	Bit32u bps;
	Bit8u bytesize;
	Bit8u stopbits;
	const char* parity;
	CSerial* cser;

	Bit16u addresses[] = {0,0x3f8,0x2f8,0x3e8,0x2e8}; // COM1,2,3,4
	Bit8u defaultirq[] = {0,4,3,4,3};

	Section_prop * section=static_cast<Section_prop *>(sec);
	
	// do not install disabled ports
	if(!section->Get_bool("enabled")) return;

	// get serialport interrupt
	irq = section->Get_int("irq");
	if(irq==2||(irq>15))			// range checking
								// (I hope irq2 is reserved, not irq9)
	{
		irq = defaultirq[comnumber];
		LOG_UART("COM%d: Invalid IRQ number. Defaulting to %d", comnumber,irq);
	}

	bps = section->Get_int("startbps");
	bytesize = section->Get_int("bytesize");
	stopbits = section->Get_int("stopbits");
	parity = section->Get_string("parity");

	// evaluate type
	typestring = section->Get_string("type");

	
	if(!stricmp(typestring, "dummy"))
	{
		cser = new CSerialDummy(
			addresses[comnumber],
			irq,bps,bytesize,parity,stopbits);
	}
	#if C_MODEM
	else if(!stricmp(typestring, "modem"))
	{
		cser = new CSerialModem(
			addresses[comnumber],
			irq,bps,bytesize,parity,stopbits,
			section->Get_string("remote"),
			section->Get_int("listenport"));
	}
	#endif // C_MODEM
	
#if C_DIRECTSERIAL
	else if(!stricmp(typestring, "directserial"))
	{
		cser = new CDirectSerial(
			 addresses[comnumber],
			 irq, bps,bytesize,parity,stopbits,
			 (char *)section->Get_string("realport"));
	}

	#endif // C_DIRECTSERIAL

	else return;

	seriallist.push_back(cser);
}

void SERIAL_Init1(Section* sec)
{
	initcomport(sec, 1);
}
void SERIAL_Init2(Section* sec)
{
	initcomport(sec, 2);
}
void SERIAL_Init3(Section* sec)
{
	initcomport(sec, 3);
}
void SERIAL_Init4(Section* sec)
{
	initcomport(sec, 4);
}